#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes import utils
from ssht00ls.classes.aliases import aliases
from ssht00ls.classes.agent import agent

# the ssync object class.
class SSync(object):
	def __init__(self):
		self.exclude = ['__pycache__', '.DS_Store']
		aliases.sync()
		self.sync()
	def mount(self, 
		# the local path.
		path=None, 
		# the remote path.
		remote=None, 
		# ssh alias.
		alias=None,
		# forced.
		forced=False,
		# exclude.
		exclude=['__pycache__', '.DS_Store'],
		# accept new host verification keys.
		accept_new_host_keys=True,
		# log level.
		log_level=0,
	):

		# checks.
		response = r3sponse.check_parameters({
			"path:str":path,
			"remote:str":remote,
			"alias:str":alias,
			"exclude:list":exclude,
			"forced:bool":forced,
			"accept_new_host_keys:bool":accept_new_host_keys, })
		if not response.success: return response
		elif os.path.exists(path):
			return r3sponse.error_response(f"Path [{path}] already exists.")
		elif "*mounted*" in str(cache.get(id=path, group="daemons")):
			return r3sponse.error_response(f"Path [{path}] is already mounted.")
		path = utils.clean_path(path)
		remote = utils.clean_path(remote)

		# pull.
		response = self.pull(
			path=path, 
			alias=alias, 
			remote=remote, 
			exclude=exclude,
			forced=forced,
			delete=True,
			safe=False,
			log_level=log_level,)
		if not r3sponse.success: return response

		# start daemon.
		p = subprocess.Popen(["python3", SOURCE_PATH, "--daemon", f"{alias}:{remote}", path, "--non-interactive"])

		# handler.
		return r3sponse.success_response(f"Successfully mounted [{alias}:{remote}] ==> [{path}].")

		#
	def unmount(self, 
		# the local path.
		path=None, 
		# forced required.
		forced=False,
		# sudo required.
		sudo=False,
		# log level.
		log_level=0,
	):

		# checks.
		response = r3sponse.check_parameters({
			"path:str":path,
			"forced:bool":forced,
			"sudo:bool":sudo, })
		if not response.success: return response
		elif not os.path.exists(path):
			return r3sponse.error_response(f"Path [{path}] does not exist.")
		elif not os.path.isdir(path):
			return r3sponse.error_response(f"Path [{path}] is not a directory.")
		elif "*mounted*" not in cache.get(id=path, group="daemons"):
			return r3sponse.error_response(f"Path [{path}] is not mounted.")
		path = utils.clean_path(path)

		# wait for daemon stop.
		response = self.stop_daemon(path=path)
		if not success: return response

		# delete local.
		"""
		if sudo: sudo_str = "sudo "
		else: sudo_str = ""
		os.system(f"{sudo_str}rm -fr {path}")
		if os.path.exists(path):
			return r3sponse.error_response(f"Failed to unmount [{path}].")
		"""

		# handler.
		return r3sponse.success_response(f"Successfully unmounted [{path}].")

		#
	def index(self, path=None, alias=None, log_level=0, accept_new_host_keys=True):

		# checks.
		if path == None:
			return r3sponse.error_response(f"Define parameter: path.")
		path = utils.clean_path(path)

		# remote.
		if alias != None:
			
			# check alias.
			response = aliases.check(alias)
			if not response.success: return response

			# check passphrase.
			if CONFIG["aliases"][alias]["smart_card"] in [True, "true", "True"]:
				response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
			else:
				response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
			if not response.success: return response
			passphrase = response.decrypted.decode()
			
			# tests.
			response = agent.add(path=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
			if not response["success"]: return response
			response = utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
			if not response.success: return response
			response = utils.test_ssh_path(alias=alias, path=path, accept_new_host_keys=accept_new_host_keys)
			if not response.success: return response

			# index.
			output = syst3m.utils.__execute_script__(f"printf 'yes' | ssh {alias} ' python3 /usr/local/lib/ssht00ls --index {path} --non-interactive --json ' ")
			try: response = json.loads(output)
			except Exception as e: 
				return r3sponse.error_response(f"Failed to serialize remote {alias} output: {output}")
			print(response["index"]["/tmp/testmount/crypto/.push_version"])
			quit()
			return r3sponse.safe_response(response)

		# local.
		else:
			if not os.path.exists(path):
				return r3sponse.error_response(f"Path [{path}] does not exist.")
			elif not os.path.isdir(path):
				return r3sponse.error_response(f"Path [{path}] is not a directory.")

			# index.
			index = {}
			for path in Files.Directory(path=path).paths(recursive=True):
				index[path] = Formats.FilePath(path).mtime(format="seconds")

			# handler.
			return r3sponse.success_response(f"Successfully indexed {len(index)} files from directory [{path}].", {
				"index":index,
			})

			#
	# pull & push.
	def pull(self,
		# the local path.
		path=None, 
		# the ssht00ls alias.
		alias=None, 
		# the remote path.
		remote=None, 
		# exlude subpaths (list) (leave None to use default).
		exclude=None,
		# update deleted files.
		delete=False,
		# forced mode.
		forced=False,
		# version control.
		safe=False,
		# accept new hosts keys.
		accept_new_host_keys=True,
		# log level.
		log_level=0,
	):	
		# check alias.
		path = utils.clean_path(path)
		remote = utils.clean_path(remote)
		response = aliases.check(alias)
		if not response.success: return response
		
		# check passphrase.
		if CONFIG["aliases"][alias]["smart_card"] in [True, "true", "True"]:
			response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
		else:
			response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
		if not response.success: return response
		passphrase = response.decrypted.decode()
		
		# tests.
		response = agent.add(path=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
		if not response["success"]: return response
		response = utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response
		response = utils.test_ssh_path(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response

		# pull.
		if exclude == None: exclude = self.exclude
		exclude_str = ""
		for i in exclude: exclude_str += f"--exclude {i} "
		delete_str = Formats.Boolean(delete).convert(true="--delete", false="")
		command = f"rsync -azP {alias}:{remote} {path} {exclude_str}{delete_str}"
		loader = syst3m.console.Loader(f"Pulling [{alias}:{remote}] to [{path}]")
		output = syst3m.utils.__execute_script__(command)
		if len(output) > 0 and output[len(output)-1] == "\n": output = output[:-1]
		cache.set(id=path, data="*mounted*", group="daemons")
		
		# handler.
		if "rsync: " in output or "rsync error: " in output:
			loader.stop(success=False)
			print(output)
			return r3sponse.success_response(f"Failed to pull [{alias}:{remote}] to [{path}].")
		else:
			loader.stop()
			return r3sponse.success_response(f"Successfully pulled [{alias}:{remote}] to [{path}].")

		#
	def push(self,
		# the local path.
		path=None, 
		# the ssht00ls alias.
		alias=None, 
		# the remote path.
		remote=None, 
		# exlude subpaths (list) (leave None to use default).
		exclude=None,
		# update deleted files.
		delete=False,
		# forced mode.
		forced=False,
		# version control.
		safe=False,
		# accept new hosts keys.
		accept_new_host_keys=True,
		# log level.
		log_level=0,
	):
		# check alias.
		path = utils.clean_path(path)
		remote = utils.clean_path(remote)
		response = aliases.check(alias)
		if not response.success: return response
		
		# check passphrase.
		if CONFIG["aliases"][alias]["smart_card"] in [True, "true", "True"]:
			response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
		else:
			response = ENCRYPTION.decrypt(CONFIG["aliases"][alias]["passphrase"])
		if not response.success: return response
		passphrase = response.decrypted.decode()
		
		# tests.
		response = agent.add(path=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
		if not response["success"]: return response
		response = utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response
		response = utils.test_ssh_path(alias=alias, path=Formats.FilePath(remote).base(), accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response

		# push.
		if exclude == None: exclude = self.exclude
		exclude_str = ""
		for i in exclude: exclude_str += f"--exclude {i} "
		delete_str = Formats.Boolean(delete).convert(true="--delete", false="")
		command = f"rsync -azP {path} {alias}:{remote} {exclude_str}{delete_str}"
		loader = syst3m.console.Loader(f"Pushing [{path}] to [{alias}:{remote}].")
		output = syst3m.utils.__execute_script__(command)
		if len(output) > 0 and output[len(output)-1] == "\n": output = output[:-1]
		cache.set(id=path, data="*mounted*", group="daemons")
		
		# handler.
		if "rsync: " in output or "rsync error: " in output:
			loader.stop(success=False)
			print(output)
			return r3sponse.success_response(f"Failed to push [{path}] to [{alias}:{remote}].")
		else:
			loader.stop()
			return r3sponse.success_response(f"Successfully pushed [{path}] to [{alias}:{remote}].")


		#
	# the mounted daemon.
	def sync(self):
		for path in Files.Directory(path=f"{cache.path}/daemons/").paths(recursive=False):
			a=1
	def daemon(self, 
		# the ssh alias.
		alias=None, 
		# the remote path.
		remote=None, 
		# thel local path.
		path=None, 
		# settings.
		start=True,
	):
		path = utils.clean_path(path)
		remote = utils.clean_path(remote)
		daemon = self.Daemon({
			"alias":alias,
			"remote":remote,
			"path":path,
			"ssync":self,
		})
		if start: daemon.start()
		return daemon
	def stop_daemon(self, path, timeout=60, sleeptime=1):
		cache.set(id=path, data="*stop*", group="daemons")
		stopped = False
		for i in range(timeout/sleeptime):
			status_ = cache.get(id=path, group="daemons")
			if "*stopped*" in status_ or "*crashed*" in status_:
				stopped = True
				break
			time.sleep(sleeptime)
		if stopped:
			return r3sponse.success_response(f"Successfully stopped ssht00ls daemon [{path}].")
		else:
			return r3sponse.error_response(f"Failed to stop ssht00ls daemon [{path}].")
	class Daemon(syst3m.objects.Thread):
		def __init__(self, attributes={}):
			syst3m.objects.Thread.__init__(self)
			self.assign(attributes)
			self.path = utils.clean_path(self.path)
			self.id = f"{self.alias}:{self.remote} ==> {self.path}"
		def run(self):

			# checks.
			if not os.path.exists(self.path):
				cache.set(id=self.path, data="*crashed*", group="daemons")
				raise ValueError(f"SSHT00ls daemon ({self.id}): Path [{self.path}] does not exist.")
			elif not os.path.isdir(self.path):
				cache.set(id=self.path, data="*crashed*", group="daemons")
				raise ValueError(f"SSHT00ls daemon ({self.id}): Path [{self.path}] is not a directory.")
			elif "*mounted*" not in cache.get(id=path, group="daemons"):
				cache.set(id=self.path, data="*crashed*", group="daemons")
				raise ValueError(f"SSHT00ls daemon ({self.id}): Path [{self.path}] is not mounted.")

			# check alias.
			response = aliases.check(alias)
			if not response.success: 
				cache.set(id=self.path, data="*crashed*", group="daemons")
				raise ValueError(response.error)

			# start.
			while True:

				# check stop command.
				if "*stop*" in cache.get(id=self.path, group="daemons"):
					break

				# if file no longer exists stop daemon.
				if not os.path.exists(self.path):
					cache.set(id=self.path, data="*crashed*", group="daemons")
					raise ValueError(f"SSHT00ls daemon ({self.id}): Mounted directory [{self.path}] from [{self.alias}:{self.remote}] no longer exists.")

			# stop.
			cache.set(id=self.path, data="*stopped*", group="daemons")

# initialized objects.
ssync = SSync()

""" """
path = "/tmp/testmount"
alias = "dev.vandenberghinc.com"
response = ssync.index(path=path, alias=alias)
print(response.index["/tmp/testmount/crypto/.push_version"])
quit()
remote_index = response["index"]
response = ssync.index(path=path)
index = response["index"]
print("Local:",index["/tmp/testmount/crypto/crypto/requirements/requirements.txt"])
print("Remote:",remote_index["/tmp/testmount/crypto/crypto/requirements/requirements.txt"])
quit()

#print(ssync.mount(
#	alias="dev.vandenberghinc.com",
#	remote="/tmp/testmount",
#	path="/tmp/testmount",))
#quit()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# syst3m.
try: import syst3m
except ImportError: 
	if os.path.exists("/usr/bin/pip3"): os.system(f"/usr/bin/pip3 install syst3m --user {OWNER}")
	else: os.system(f"pip3 install syst3m")

# updates.
try: import cl1
except ImportError: 
	if os.path.exists("/usr/bin/pip3"): os.system(f"/usr/bin/pip3 install cl1 --user {OWNER}")
	else: os.system(f"pip3 install cl1")
if cl1.arguments_present(["--update"]):
	os.system(f"curl https://raw.githubusercontent.com/vandenberghinc/{ALIAS}/master/{ALIAS}/requirements/installer.remote | bash ")
	sys.exit(0)

# imports.
try: 

	# imports.
	import os, sys, requests, ast, json, pathlib, glob, platform, subprocess, pexpect, random, getpass, time

	# inc imports.
	from fil3s import Files, Formats
	from r3sponse import r3sponse
	import cl1, syst3m, encrypti0n, netw0rk

# download.
except ImportError as e:
	import os
	if os.path.exists("/usr/bin/pip3"): 
		os.system(f"/usr/bin/pip3 install -r {syst3m.defaults.get_source_path(__file__, back=3)}/requirements/requirements.txt --user {os.environ.get('USER')}")
	else:
		os.system(f"pip3 install -r {syst3m.defaults.get_source_path(__file__, back=3)}/requirements/requirements.txt")

	# imports.
	import os, sys, requests, ast, json, pathlib, glob, platform, subprocess, pexpect, random, getpass, time

	# inc imports.
	from fil3s import Files, Formats
	from r3sponse import r3sponse
	import cl1, syst3m, encrypti0n, netw0rk


# source.	
ALIAS = "ssht00ls"
SOURCE_NAME = "ssht00ls"
SOURCE_PATH = syst3m.defaults.get_source_path(__file__, back=3)
BASE = syst3m.defaults.get_source_path(SOURCE_PATH)
OS = syst3m.defaults.check_operating_system(supported=["linux", "osx"])
syst3m.defaults.check_alias(alias=ALIAS, executable=f"{SOURCE_PATH}")

# universal variables.
OWNER = os.environ.get("USER")
GROUP = "root"
HOME_BASE = "/home/"
HOME = f"/home/{os.environ.get('USER')}/"
MEDIA = f"/media/{os.environ.get('USER')}/"
if OS in ["osx"]: 
	HOME_BASE = "/Users/"
	HOME = f"/Users/{os.environ.get('USER')}/"
	MEDIA = f"/Volumes/"
	GROUP = "wheel"

# network.
NETWORK_INFO = netw0rk.network.info()
if not NETWORK_INFO.success: 
	r3sponse.log(error=NETWORK_INFO.error, json=cl1.arguments_present(["--json", "-j"]))
	sys.exit(1)

# check lib.
if not os.path.exists(f"{SOURCE_PATH}/lib"):
	print("Downloading ssht00ls library.")
	os.system(f"rm -fr /tmp/ssht00ls && git clone https://github.com/vandenberghinc/ssht00ls /tmp/ssht00ls && cp -r /tmp/ssht00ls/ssht00ls/lib {SOURCE_PATH}/lib")

# check usr lib.
if not os.path.exists("/usr/local/lib/ssht00ls"):
	print("Installing ssht00ls into system library.")
	os.system(f"rsync -azP {SOURCE_PATH}/ /usr/local/lib/ssht00ls")

# database.
for dir, permission in [
	[f"{HOME}/.{ALIAS}", 770],
	[f"{HOME}/.{ALIAS}/lib", 770],
	[f"{HOME}/.{ALIAS}/.cache", 770],
]:
	if not os.path.exists(dir): os.system(f"sudo mkdir {dir} && sudo chown {OWNER}:{GROUP} {dir} && sudo chmod {permission} {dir}")

# files.
DATABASE = Files.Directory(path=f"{HOME}/.{ALIAS}")
CONFIG = Files.Dictionary(path=syst3m.env.fill(syst3m.env.get_string("SSHT00LS_CONFIG", default=DATABASE.join("config",""))), load=True, default={
	"aliases":{
		"example.com (key's are optional)":{
			"user":"administrator",
			"public_ip":"192.168.1.100",
			"public_port":22,
			"private_ip":"84.84.123.192",
			"private_port":22,
			"private_key":"~/keys/example.com/administrator/private_key",
			"public_key":"~/keys/example.com/administrator/public_key",
			"passphrase":None,
			"smart_card":False,
			"pin":None,
		}
	},
	"settings": {
		"keep_alive":60,
	},
	"encryption": {
		"public_key":None,
		"private_key":None,
	},
})

# universal options.
INTERACTIVE = not cl1.arguments_present(["--non-interactive"])
JSON = cl1.arguments_present(["-j", "--json"])

# initialize cache.
cache = syst3m.cache.Cache(
	path=f"{HOME}/.{ALIAS}/.cache/",)
webserver = syst3m.cache.WebServer(
	path=f"{HOME}/.{ALIAS}/.cache/",
	host="127.0.0.1",
	port=52379,)
if cl1.argument_present("--stop-agent"):
	if not webserver.running(): 
		r3sponse.log(error="The ssht00ls agent is not running.", json=JSON)
		sys.exit(1)
	processes = syst3m.processes(includes="ssht00ls/ --start-agent")
	if not processes.success: 
		r3sponse.log(response=processes, json=json)
		sys.exit(0)
	if len(processes.processes) <= 1:
		r3sponse.log(error="Unable to find the pid of the ssht00ls agent.", json=json)
		sys.exit(0)
	for pid, info in processes.processes.items():
		if info["process"] not in ["grep"]:
			response = syst3m.kill(pid=pid)
			if not response.success: 
				r3sponse.log(response=processes, json=json)
				sys.exit(0)
	r3sponse.log(message="Successfully stopped the ssht00ls agent.", json=json)
	sys.exit(0)
elif cl1.argument_present("--start-agent"):
	if not webserver.running(): 
		webserver.start()
		sys.exit(0)
	else:
		r3sponse.log(error="The ssht00ls agent is already running.", json=JSON)
		sys.exit(1)
else:
	if not INTERACTIVE and not webserver.running(): 
		print("Starting the ssht00ls agent.")
		p = subprocess.Popen(["python3", SOURCE_PATH, "--start-agent", "2>", "/dev/null"])
		time.sleep(3)

# encryption.
passphrase = None
if None in [CONFIG.dictionary["encryption"]["public_key"], CONFIG.dictionary["encryption"]["private_key"]]:
	if INTERACTIVE:
		print(f"INTERACTIVE 1: {INTERACTIVE}")
		passphrase = getpass.getpass("Enter the passphrase of the ssht00ls encryption:")
		if len(passphrase) < 8: 
			r3sponse.log(error="The passphrase must contain at least 8 characters.", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)
		elif passphrase.lower() == passphrase: 
			r3sponse.log(error="The passphrase must contain capital characters.", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)
		elif passphrase != getpass.getpass("Enter the same passphrase:"): 
			r3sponse.log(error="The passphrase must contain at least 8 characters.", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)
		ENCRYPTION = encrypti0n.aes.AsymmetricAES(
			public_key=CONFIG.dictionary["encryption"]["public_key"],
			private_key=CONFIG.dictionary["encryption"]["public_key"],
			passphrase=passphrase,
			memory=True,)
		response = ENCRYPTION.generate_keys()
		if not response["success"]: 
			r3sponse.log(error=f"Encoutered an error while generating the master encryption key: {response['error']}", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)
		ENCRYPTION.rsa.private_key = response.private_key
		ENCRYPTION.rsa.public_key = response.public_key
		CONFIG.dictionary["encryption"]["public_key"] = ENCRYPTION.rsa.public_key
		CONFIG.dictionary["encryption"]["private_key"] = ENCRYPTION.rsa.private_key
		CONFIG.save()
		response = ENCRYPTION.load_keys()
		if not response["success"]: 
			r3sponse.log(error=f"Encoutered an error while activating the ssht00ls encryption: {response['error']}", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)
		response = webserver.set(group="passphrases", id="master", data=passphrase)
		if not response["success"]: 
			r3sponse.log(error=f"Encoutered an error while parsing the webserver: {response['error']}", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)
	else:
		r3sponse.log(error="There is no encryption installed.", json=JSON)
		sys.exit(1)
else:
	response, new, passphrase = webserver.get(group="passphrases", id="master"), False, None
	if response["success"]:
		passphrase = response["data"]
	if passphrase == None:
		if not INTERACTIVE:
			r3sponse.log(error=response.error, json=JSON)
			sys.exit(1)
		else:
			new = True
			print(f"INTERACTIVE 2: {INTERACTIVE}")
			passphrase = getpass.getpass("Enter the passphrase of the ssht00ls encryption:")
	ENCRYPTION = encrypti0n.aes.AsymmetricAES(
		public_key=CONFIG.dictionary["encryption"]["public_key"],
		private_key=CONFIG.dictionary["encryption"]["private_key"],
		passphrase=passphrase,
		memory=True,)
	response = ENCRYPTION.load_keys()
	if not response["success"]: 
		r3sponse.log(error=f"Encoutered an error while activating the ssht00ls encryption: {response['error']}", json=cl1.arguments_present(["--json", "-j"]))
		sys.exit(1)
	if new:
		response = webserver.set(group="passphrases", id="master", data=passphrase)
		if not response["success"]: 
			r3sponse.log(error=f"Encoutered an error while parsing the webserver: {response['error']}", json=cl1.arguments_present(["--json", "-j"]))
			sys.exit(1)

# encrypted database.
if passphrase != None:
	ENCRYPTED_DATABASE = encrypti0n.aes.Database(path=f"{HOME}/.{ALIAS}/.cache.enc/", aes=ENCRYPTION)
	response = ENCRYPTED_DATABASE.activate()
	if not response["success"]: 
		r3sponse.log(error=f"Encoutered an error while parsing the webserver: {response['error']}", json=cl1.arguments_present(["--json", "-j"]))
		sys.exit(1)
	PASSPHRASES = ENCRYPTED_DATABASE.load("passphrases")

	#